/***************************************************************************************

***************************************************************************************/


/*
*************************************************************************************
* File name: image.c
* Function: Decode a Slice
*
*************************************************************************************
*/


#include "contributors.h"

#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
#include <string.h>
#include <assert.h>
#ifdef WIN32
#include <IO.H>
#include <STDIO.H>
#endif

#include "global.h"
#include "mbuffer.h"
#include "header.h"
#include "annexb.h"
#include "vlc.h"
#include "memalloc.h"
#include "bbv.h" 

//!EDIT START <added by  AEC
#include "context_ini.h"    
#include "AEC.h"   
#include "biaridecod.h"
//!EDIT end <added by  AEC

static void rotate_buffer ();
void copy2buffer(struct img_par *img, struct inp_par *inp,int bot);
void replace2buffer(struct img_par *img, struct inp_par *inp,int bot);
void DecideMvRange();	// ITM

BbvBuffer_t *pBbv = NULL; 


unsigned char *temp_slice_buf;
int first_slice_length;
int first_slice_startpos;

extern StatBits *StatBitsPtr;	

/*
*************************************************************************
* Function:decodes one I- or P-frame
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
int decode_one_frame(struct img_par *img,struct inp_par *inp, struct snr_par *snr)
{
  int current_header;

  time_t ltime1;                  // for time measurement
  time_t ltime2;

#ifdef WIN32
  struct _timeb tstruct1;
  struct _timeb tstruct2;
#else
  struct timeb tstruct1;
  struct timeb tstruct2;
#endif

  int tmp_time;                   // time used by decoding the last frame

  float framerate[8]={24000/1001,24,25,30000/1001,30,50,60000/1001,60};  //ITM_r2

#ifdef WIN32
  _ftime (&tstruct1);             // start time ms
#else
  ftime (&tstruct1);              // start time ms
#endif
  time( &ltime1 );                // start time s

  img->current_mb_nr = -4711; // initialized to an impossible value for debugging -- correct value is taken from slice header

  if (input->check_BBV_flag && img->number && pBbv)
  {
    pBbv->frame_code_bits = 0;
  }
  current_header = Header();

  DecideMvRange();  //ITM

  if (current_header == EOS)
    return EOS;

  {
    if (horizontal_size% 16 != 0)   
    {
      img->auto_crop_right = 16-(horizontal_size % 16);
    }
    else
    {
      img->auto_crop_right=0;
    }
    if (!progressive_sequence) 
    {
      if (vertical_size % 32 != 0)
      {
        img->auto_crop_bottom = 32-(vertical_size % 32);
      }
      else
      {
        img->auto_crop_bottom=0;
      }
    }
    else              
    {
      if (vertical_size % 16 != 0)   
      {
        img->auto_crop_bottom = 16-(vertical_size % 16);
      }
      else
      {
        img->auto_crop_bottom=0;
      }
    }                                 
    // Reinit parameters (NOTE: need to do before init_frame // 
    img->width          = (horizontal_size+img->auto_crop_right);
    img->height         = (vertical_size+img->auto_crop_bottom);
    img->width_cr       = (img->width>>1);
    img->height_cr      = (img->height>>1);
    img->PicWidthInMbs  = img->width/MB_BLOCK_SIZE;
    img->PicHeightInMbs = img->height/MB_BLOCK_SIZE;
    img->PicSizeInMbs   = img->PicWidthInMbs * img->PicHeightInMbs;
    img->max_mb_nr      =(img->width * img->height) / (MB_BLOCK_SIZE * MB_BLOCK_SIZE);

  }

  img->current_mb_nr = 0;

  init_frame(img, inp);

	img->types = img->type;   

	if((img->type==P_IMG)&&((pre_dec_img_type == I_IMG)||(pre_dec_img_type==B_IMG)))
	  b_pre_dec_intra_img=TRUE;
	else if ((img->type==P_IMG)&&(pre_dec_img_type==P_IMG))
	  b_pre_dec_intra_img=FALSE; 

	pre_dec_img_type=img->type;	
	if(img->type!=B_IMG)
	{		
	  pre_img_type  = img->type;		
	  pre_img_types = img->types;		
	}

    picture_data(img, inp);

    imgY = imgY_frm;
    imgUV = imgUV_frm;

  img->height = (vertical_size+img->auto_crop_bottom);
  img->height_cr = /*vertical_size*/ img->height/2;	
  img->PicWidthInMbs  = img->width/MB_BLOCK_SIZE;
  img->PicHeightInMbs = img->height/MB_BLOCK_SIZE;
  img->PicSizeInMbs   = img->PicWidthInMbs * img->PicHeightInMbs;


  frame_postprocessing(img);


  if (img->type==B_IMG) {
    if (p_ref)
      find_snr(snr,img,p_ref);      // if ref sequence exist
  }



#ifdef WIN32
  _ftime (&tstruct2);   // end time ms
#else
  ftime (&tstruct2);    // end time ms
#endif

  time( &ltime2 );                                // end time sec
  tmp_time=(int)((ltime2*1000+tstruct2.millitm) - (ltime1*1000+tstruct1.millitm));
  tot_time=tot_time + tmp_time;

  //ITM_r2
  StatBitsPtr->curr_frame_bits = StatBitsPtr->curr_frame_bits*8 + StatBitsPtr->emulate_bits - StatBitsPtr->last_unit_bits;
  StatBitsPtr->bitrate += StatBitsPtr->curr_frame_bits;
  StatBitsPtr->coded_pic_num++;
  if( (int)(StatBitsPtr->coded_pic_num - (StatBitsPtr->time_s+1)*framerate[frame_rate_code-1] + 0.5) == 0 )
  {
    StatBitsPtr->total_bitrate[StatBitsPtr->time_s++] = StatBitsPtr->bitrate;
    CheckBitrate(StatBitsPtr->bitrate,StatBitsPtr->time_s);
    StatBitsPtr->bitrate = 0;
  }

  if (img->type!=B_IMG) {
    pre_tmp_time = tmp_time;
    pre_img_tr   = img->tr;
    pre_img_qp   = img->qp;
    StatBitsPtr->prev_frame_bits = StatBitsPtr->curr_frame_bits; 
    StatBitsPtr->prev_emulate_bits = StatBitsPtr->emulate_bits;  
  }	 

  if(input->ref_pic_order) {    //ref order
    if(img->type == I_IMG) // I picture
      fprintf(stdout,"%3d(I)  %3d %5d %7.4f %7.4f %7.4f %5d\n",
      dec_ref_num, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time);
    else if(img->type == P_IMG) // P pictures
    {
      if (img->types == I_IMG)
        fprintf(stdout,"%3d(I)  %3d %5d %7.4f %7.4f %7.4f %5d\n",
        dec_ref_num, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time);  
      else
        fprintf(stdout,"%3d(P)  %3d %5d %7.4f %7.4f %7.4f %5d\n",
        dec_ref_num, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time);
    }
    else // B pictures
      fprintf(stdout,"%3d(B)  %3d %5d %7.4f %7.4f %7.4f %5d\n",
      dec_ref_num, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time);
  }
  else if(img->type==B_IMG)
    fprintf(stdout,"%3d(B)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s %8d %6d\n",
    FrameNum, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time,"FRM",StatBitsPtr->curr_frame_bits,StatBitsPtr->emulate_bits);

  if (input->check_BBV_flag)
    pBbv->frame_code_bits        = StatBitsPtr->curr_frame_bits;     
  StatBitsPtr->curr_frame_bits = 0;								
  StatBitsPtr->emulate_bits    = 0;								
  StatBitsPtr->last_unit_bits  = 0;								

  fflush(stdout);

  if(img->type==B_IMG)
    report_frame(snr,tmp_time);

  if(img->type == I_IMG || img->type == P_IMG) // I or P pictures
  {
    copy_Pframe(img);  
    Update_Picture_Buffers();
  }
  else if (input->output_dec_pic)  // output_dec_pic // B pictures
    write_frame(img,p_out);  // write image to output YUV file 

  if (input->check_BBV_flag)
  {    
    update_bbv(pBbv, pBbv->frame_code_bits);
  }


  if(img->type == I_IMG || img->type == P_IMG)   // I or P pictures
    img->number++;
  else
    Bframe_ctr++;    // B pictures

  return (SOP);
}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
void report_frame(struct snr_par    *snr,int tmp_time)
{
  FILE *file;
  file = fopen("stat.dat","at");

  if(img->type == I_IMG) // I picture
    fprintf(file,"%3d(I)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s\n",
    FrameNum, pre_img_tr, pre_img_qp,snr->snr_y,snr->snr_u,snr->snr_v,pre_tmp_time,"FRM"); 

  else if(img->type == P_IMG) // P pictures
    fprintf(file,"%3d(P)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s\n",
    FrameNum, pre_img_tr, pre_img_qp,snr->snr_y,snr->snr_u,snr->snr_v,pre_tmp_time,"FRM");

  else // B pictures
    fprintf(file,"%3d(B)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s\n",
    FrameNum, img->tr, img->qp,snr->snr_y,snr->snr_u,snr->snr_v,tmp_time,"FRM");

  FrameNum++;	

  fclose(file);
}

/*
*************************************************************************
* Function:Find PSNR for all three components.Compare decoded frame with
the original sequence. Read inp->jumpd frames to reflect frame skipping.
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

#ifdef WIN32
void find_snr(
struct snr_par *snr,   //!< pointer to snr parameters
struct img_par *img,   //!< pointer to image parameters
  int p_ref)  
#else
void find_snr(
struct snr_par *snr,   //!< pointer to snr parameters
struct img_par *img,   //!< pointer to image parameters
  FILE *p_ref)           //!< filestream to reference YUV file
#endif
{
  int i,j;
  int diff_y,diff_u,diff_v;
  int uv;
  int uvformat = 4; 
#ifdef WIN32
  __int64 status;
#else
  int  status;
#endif  

  int img_width = (img->width-img->auto_crop_right);
  int img_height = (img->height-img->auto_crop_bottom);
  int img_width_cr = (img_width/2);
  int img_height_cr = (img_height/2); 
#ifdef WIN32
  const unsigned int  bytes_y = img_width * img_height;
  const unsigned int  bytes_uv = img_width_cr * img_height_cr;  
  const __int64 framesize_in_bytes = (__int64)(bytes_y + 2*bytes_uv);
#else
  const unsigned int  bytes_y = img_width * img_height;
  const unsigned int  bytes_uv = img_width_cr * img_height_cr;
  const int framesize_in_bytes = bytes_y + 2*bytes_uv;
#endif  
  unsigned char *buf;
  int   offset_units;

  if(!input->yuv_structure)
    buf = malloc (bytes_y);
  else
    buf = malloc (bytes_y+2*bytes_uv);
  if (NULL == buf)
  {
    no_mem_exit("find_snr: buf");
  }
  snr->snr_y=0.0;
  snr->snr_u=0.0;
  snr->snr_v=0.0;

#ifndef WIN32
  rewind(p_ref);
#endif  
  if(RefPicExist)
  {
#ifdef WIN32
    if(!input->ref_pic_order)    //ref order
      status = _lseeki64 (p_ref, framesize_in_bytes * FrameNum, SEEK_SET); 
    else
      status = _lseeki64 (p_ref, framesize_in_bytes * dec_ref_num, SEEK_SET);
#else
    if(!input->ref_pic_order)    //ref order
      status = fseek (p_ref, framesize_in_bytes * FrameNum, 0);
    else
      status = fseek (p_ref, framesize_in_bytes * dec_ref_num, 0);	
#endif

#ifdef WIN32
    if (status == -1)
#else
    if (status != 0)
#endif	
    {
      snprintf(errortext, ET_SIZE, "Error in seeking img->tr: %d", img->tr);
      RefPicExist = 0;
    }
  } 

#ifdef WIN32
  if(!input->yuv_structure)
    _read(p_ref, buf, bytes_y);
  else
    _read(p_ref, buf, bytes_y+2*bytes_uv);
  if(!input->yuv_structure)  //YUV
  {
    for(j=0;j<img_height;j++)
      memcpy(&imgY_ref[j][0], buf+j*img_width, img_width);
  }
  else  //U0Y0 V1Y1
  {
    for(j=0;j<img_height;j++)
      for(i=0;i<img_width;i++)
      {
        offset_units = j*img_width + i;
        imgY_ref[j][i] = buf[2*offset_units+1];
        if(offset_units%2==0) //U component
          imgUV_ref[0][j][i/2] = buf[2*offset_units];
        else                  //V component
          imgUV_ref[1][j][i/2] = buf[2*offset_units];
      }
  }
#else
  if(!input->yuv_structure)
  {
    for (j=0; j< img_height; j++)
      for(i=0; i< img_width; i++)
        imgY_ref[j][i]=fgetc(p_ref);
  }
  else
  {
    for (j=0; j< img_height; j++)
      for(i=0; i< img_width; i++)
      {
        offset_units = j*img_width + i;
        if(offset_units%2==0) //U component
          imgUV_ref[0][j][i/2] = fgetc(p_ref);
        else				  //V component
          imgUV_ref[1][j][i/2] = fgetc(p_ref);

        imgY_ref[j][i] = fgetc(p_ref);
      }
  }
#endif      

  for (uv=0; uv < 2; uv++)
  {	  
    if(!input->yuv_structure)
    {
#ifdef WIN32
      _read(p_ref, buf, img_height_cr * img_width_cr);
      for(j=0;j<img_height_cr;j++)
        memcpy(&imgUV_ref[uv][j][0], buf+j*img_width_cr, img_width_cr);
#else
      for(j=0;j<img_height_cr;j++)
        for (i=0; i < img_width_cr; i++)
          imgUV_ref[uv][j][i]=fgetc(p_ref);
#endif
    }
  }

  img->quad[0]=0;
  diff_y=0;
  for (j=0; j < img_height; ++j)
  {
    for (i=0; i < img_width; ++i)
    {
      if(img->type==B_IMG && !eos || input->ref_pic_order)	
        diff_y += img->quad[abs(imgY[j][i]-imgY_ref[j][i])];
      else
        diff_y += img->quad[abs(imgY_prev[j][i]-imgY_ref[j][i])];

      if(img->type==B_IMG && !eos)
      {
        if(imgY[j][i]!=imgY_ref[j][i])
        {
          int mb_nr;
          mb_nr = (j/16)*(img->width/16)+i/16;
        }
      }
      else
      {
        if(/*imgY*/imgY_ref[j][i]!=imgY_prev[j][i]) 
        {
          int mb_nr;
          mb_nr = (j/16)*(img->width/16)+i/16;
		  mb_nr = mb_nr;
        }
      }
    }
  }

  // Chroma
  diff_u=0;
  diff_v=0;

  for (j=0; j < img_height_cr; ++j)
  {
    for (i=0; i < img_width_cr; ++i)
    {
      if(img->type==B_IMG && !eos || input->ref_pic_order){  //ref order
		diff_u += img->quad[abs(imgUV_ref[0][j][i]-imgUV[0][j][i])];
		diff_v += img->quad[abs(imgUV_ref[1][j][i]-imgUV[1][j][i])];
      }
      else{
		diff_u += img->quad[abs(imgUV_ref[0][j][i]-imgUV_prev[0][j][i])];
		diff_v += img->quad[abs(imgUV_ref[1][j][i]-imgUV_prev[1][j][i])];
      }

      if(img->type==B_IMG && !eos)
      {
        if(imgUV_ref[0][j][i]!=imgUV[0][j][i])
        {
          int mb_nr;
          mb_nr = (j/16)*(img->width/16)+i/8;
        }
      }
      else
      {
        if(imgUV_prev[0][j][i]!=imgUV[0][j][i])
        {
          int mb_nr;
          mb_nr = (j/16)*(img->width/16)+i/8;
        }
      }
    }
  }

  // Collecting SNR statistics
  if (diff_y != 0)
    snr->snr_y=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)diff_y));        // luma snr for current frame
  if (diff_u != 0)
    snr->snr_u=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)(/*4*/uvformat*diff_u)));    //  chroma snr for current frame
  if (diff_v != 0)
    snr->snr_v=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)(/*4*/uvformat*diff_v)));    //  chroma snr for current frame,

  if (img->number == /*0*/1) // first
  {
    snr->snr_y1=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)diff_y));       // keep luma snr for first frame
    snr->snr_u1=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)(/*4*/uvformat*diff_u)));   // keep chroma snr for first frame,
    snr->snr_v1=(float)(10*log10(65025*(float)(img_width/*img->width*/)*(img_height/*img->height*/)/(float)(/*4*/uvformat*diff_v)));   // keep chroma snr for first frame,
    snr->snr_ya=snr->snr_y1;
    snr->snr_ua=snr->snr_u1;
    snr->snr_va=snr->snr_v1;

    if (diff_y == 0)   
      snr->snr_ya=/*50*/0;
    if (diff_u == 0)
      snr->snr_ua=/*50*/0;
    if (diff_v == 0)
      snr->snr_va=/*50*/0;
  }
  else
  {
    snr->snr_ya=(float)(snr->snr_ya*(img->number+Bframe_ctr-1)+snr->snr_y)/(img->number+Bframe_ctr); // average snr chroma for all frames 
    snr->snr_ua=(float)(snr->snr_ua*(img->number+Bframe_ctr-1)+snr->snr_u)/(img->number+Bframe_ctr); // average snr luma for all frames 
    snr->snr_va=(float)(snr->snr_va*(img->number+Bframe_ctr-1)+snr->snr_v)/(img->number+Bframe_ctr); // average snr luma for all frames 
  }

  free (buf);
}


/*
*************************************************************************
* Function:Interpolation of 1/4 subpixel
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
#ifdef IVC_PROFILE

void get_block_adaptive(int ref_frame,int x_pos, int y_pos, struct img_par *img, int block[8][8], unsigned char **ref_pic)
{
    int dx, dy;
    int x, y;
    int i, j;
    int maxold_x,maxold_y;
    int result;
    int tmp_res[26][26];

    int tap, tap_min, tap_max;

    static const int COEF_4tap[3][4] =
    {
        { -6, 56, 15, -1 },
        { -4, 36, 36, -4 },
        { -1, 15, 56, -6 }
    };

    static const int COEF_6tap[3][6] =
    {
        { 2, -9, 57, 17, -4, 1 },
        { 2, -9, 39, 39, -9, 2 },
        { 1, -4, 17, 57, -9, 2 }
    };

    static const int COEF_8tap[3][8] =
    {
        { -1, 4, -10, 57, 19,  -7, 3, -1 },
        { -1, 4, -11, 40, 40, -11, 4, -1 },
        { -1, 3,  -7, 19, 57, -10, 4, -1 }
    };

    static const int COEF_10tap[3][10] =
    {
        { 1, -2, 4, -10, 57, 19,  -7, 3, -1, 0 },
        { 1, -2, 5, -12, 40, 40, -12, 5, -2, 1 },
        { 0, -1, 3,  -7, 19, 57, -10, 4, -2, 1 }
    };

    const int *COEF[3];

    if(img->height >= 1600)//for UHD, 2560x1600
    {
        COEF[0] = COEF_4tap[0];
        COEF[1] = COEF_4tap[1];
        COEF[2] = COEF_4tap[2];
        tap = 4;
    }
    else if(img->height >= 720)//for 720p and 1080p
    {
        COEF[0] = COEF_6tap[0];
        COEF[1] = COEF_6tap[1];
        COEF[2] = COEF_6tap[2];
        tap = 6;
    }
    else
    {
        COEF[0] = COEF_10tap[0];
        COEF[1] = COEF_10tap[1];
        COEF[2] = COEF_10tap[2];
        tap = 10;
    }

    tap_min = (tap>>1) -1;
    tap_max = (tap>>1) +1;

    //    A  a  b  c  B
    //    d  e  f  g
    //    h  i  j  k
    //    n  p  q  r
    //    C           D

    dx = x_pos&3;
    dy = y_pos&3;
    x_pos = (x_pos-dx)/4;
    y_pos = (y_pos-dy)/4;
    maxold_x = img->width-1;
    maxold_y = img->height-1;

    if (dx == 0 && dy == 0) {  //fullpel position: A
        for (j = 0; j < B8_SIZE; j++)
            for (i = 0; i < B8_SIZE; i++)
                block[i][j] = ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i))];
    }
    else 
    { /* other positions */
        if(dy==0){//horizonal  position: a,b,c
            for (j = 0; j < B8_SIZE; j++){
                for (i = 0; i < B8_SIZE; i++) {
                    for ( result = 0, x = -tap_min; x < tap_max; x++ )
                        result += ref_pic[max ( 0, min ( maxold_y, y_pos + j ) ) ][max ( 0, min ( maxold_x, x_pos + i + x ) ) ] * COEF[dx - 1][x + tap_min];
                    block[i][j] = max(0, min(255, (result+32)/64));
                }
            }
        }
        else if(dx==0){//vertical position: d,h,n
            for (j = 0; j < B8_SIZE; j++) {
                for (i = 0; i < B8_SIZE; i++) {
                    for ( result = 0, y = -tap_min; y < tap_max; y++ )
                        result += ref_pic[max ( 0, min ( maxold_y, y_pos + j + y ) ) ][max ( 0, min ( maxold_x, x_pos + i ) ) ] * COEF[dy - 1][y + tap_min];
                    block[i][j] = max(0, min(255, (result+32)/64));
                }
            }
        }
        else{
            //for (j = -4; j < B8_SIZE+5; j++) {//bug,hlv
            for (j = -tap_min; j < B8_SIZE + (tap_max - 1); j++) {
                for (i = 0; i < B8_SIZE; i++) {
                    for ( result = 0, x = -tap_min; x < tap_max; x++ )
                        result += ref_pic[max ( 0, min ( maxold_y, y_pos + j ) ) ][max ( 0, min ( maxold_x, x_pos + i + x ) ) ] * COEF[dx - 1][x + tap_min];
                    tmp_res[j + tap_min][i] = result;
                }
            }
            for (j = 0; j < B8_SIZE; j++) {
                for (i = 0; i < B8_SIZE; i++) {
                    for ( result = 0, y = -tap_min; y < tap_max; y++ )
                        result += tmp_res[j + tap_min + y][i] * COEF[dy - 1][y + tap_min];
                    block[i][j] = max(0, min(255, (result+2048)/4096));
                }
            }
        }
    }
}


void get_block(int ref_frame,int x_pos, int y_pos, struct img_par *img, int block[8][8], unsigned char **ref_pic)
{
	int dx, dy;
	int x, y;
	int i, j;
	int maxold_x,maxold_y;
	int result;
	int tmp_res[26][26];
	static const int COEF_HALF_8tap[8] = {
		-1, 4, -11, 40, 40, -11, 4, -1
	};
	static const int COEF_QUART1_8tap[8] = {
		-1, 4, -10, 57, 18, -6, 3, -1
	};
	static const int COEF_QUART3_8tap[8] = {
		-1, 3, -6, 18, 57, -10, 4, -1
	};

	static const int COEF_HALF_6tap[6] = {
		2, -9, 39, 39, -9, 2
	};
	static const int COEF_QUART1_6tap[6] = {
		2, -9, 57, 17, -4, 1
	};
	static const int COEF_QUART3_6tap[6] = {
		1, -4, 17, 57, -9, 2
	};

    if (if_type == 1) {
        get_block_adaptive(ref_frame,x_pos, y_pos, img, block, ref_pic);
        return;
    }
	//    A  a  b  c  B
	//    d  e  f  g
	//    h  i  j  k
	//    n  p  q  r
	//    C           D

	dx = x_pos&3;
	dy = y_pos&3;
	x_pos = (x_pos-dx)/4;
	y_pos = (y_pos-dy)/4;
	maxold_x = img->width-1;
	maxold_y = img->height-1;

	if (dx == 0 && dy == 0) {  //fullpel position: A
		for (j = 0; j < B8_SIZE; j++)
			for (i = 0; i < B8_SIZE; i++)
				block[i][j] = ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i))];
	}
	else 
	{ /* other positions */
		if((dx==2) && (dy==0)){//horizonal 1/2 position: b
			for (j = 0; j < B8_SIZE; j++){
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++) //-3~4
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_HALF_8tap[x+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		else if((dx==1) && (dy==0)){//horizonal 1/4 position: a
			for (j = 0; j < B8_SIZE; j++){
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++) //-3~4
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART1_8tap[x+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		else if((dx==3) && (dy==0)){//horizonal 3/4 position: a
			for (j = 0; j < B8_SIZE; j++){
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++) //-3~4
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART3_8tap[x+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		//vertical
		else if((dy==2) && (dx==0)){//vertical 1/2 position: h
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -3; y < 5; y++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j+y))][max(0,min(maxold_x,x_pos+i))]*COEF_HALF_8tap[y+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		else if((dy==1) && (dx==0)){//vertical 1/4 position: d
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -3; y < 5; y++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j+y))][max(0,min(maxold_x,x_pos+i))]*COEF_QUART1_8tap[y+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		else if((dy==3) && (dx==0)){//vertical 3/4 position: n
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -3; y < 5; y++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j+y))][max(0,min(maxold_x,x_pos+i))]*COEF_QUART3_8tap[y+3];
					block[i][j] = max(0, min(255, (result+32)/64));
				}
			}
		}

		//remaining 9 fractional-pel pixels
		//e,i,p
		else if((dx==1) && (dy==1)){//horizontal 1/4 and vertical 1/4 position: e
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART1_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART1_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==1) && (dy==2)){//horizontal 1/4 and vertical 1/2 position: i
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART1_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_HALF_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==1) && (dy==3)){//horizontal 1/4 and vertical 3/4 position: p
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART1_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART3_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		//f,j,q
		else if((dx==2) && (dy==1)){//horizontal 1/2 and vertical 1/4 position: f
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_HALF_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART1_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==2) && (dy==2)){//horizontal 1/4 and vertical 1/2 position: j
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_HALF_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_HALF_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==2) && (dy==3)){//horizontal 1/2 and vertical 3/4 position: q
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_HALF_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART3_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}


		//g,k,r
		else if((dx==3) && (dy==1)){//horizontal 3/4 and vertical 1/4 position: g
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART3_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART1_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==3) && (dy==2)){//horizontal 3/4 and vertical 1/2 position: k
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART3_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_HALF_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}

		else if((dx==3) && (dy==3)){//horizontal 3/4 and vertical 3/4 position: r
			for (j = -2; j < B8_SIZE+3; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, x = -3; x < 5; x++)
						result += ref_pic[max(0,min(maxold_y,y_pos+j))][max(0,min(maxold_x,x_pos+i+x))]*COEF_QUART3_8tap[x+3];
					tmp_res[j+2][i] = result;
				}
			}
			for (j = 0; j < B8_SIZE; j++) {
				for (i = 0; i < B8_SIZE; i++) {
					for (result = 0, y = -2; y < 4; y++)
						result += tmp_res[j+y+2][i]*COEF_QUART3_6tap[y+2];
					block[i][j] = max(0, min(255, (result+2048)/4096));
				}
			}
		}


	}
}
#endif

/*
*************************************************************************
* Function:Reads the content of two successive startcode from the bitstream
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

int Header()
{
  unsigned char *Buf;
  int startcodepos,length; 
  static unsigned long prev_pos=0 , curr_pos=0;	
  static unsigned long prev_pic=0, curr_pic=0;	

  if ((Buf = (char*)calloc (MAX_CODED_FRAME_SIZE , sizeof(char))) == NULL)
    no_mem_exit("GetAnnexbNALU: Buf");

  while (1)
  {     
    StartCodePosition = GetOneUnit(Buf,&startcodepos,&length);    

    switch(Buf[startcodepos])
    {
    case SEQUENCE_HEADER_CODE:
      if (input->check_BBV_flag && !img->new_seq_header_flag && pBbv)
      {
        stat_bbv_buffer(pBbv);
        pBbv = free_bbv_memory(pBbv);
        img->new_seq_header_flag = 1;
      }

      img->new_sequence_flag = 1; 
      SequenceHeader(Buf,startcodepos,length);

      if (input->check_BBV_flag && img->new_seq_header_flag)
      {
        pBbv = init_bbv_memory(frame_rate_code, low_delay, bbv_buffer_size, bit_rate_upper, bit_rate_lower);
        img->new_seq_header_flag = 0;
      }

      img->seq_header_indicate = 1;               
      break;
    case EXTENSION_START_CODE:
      extension_data(Buf,startcodepos,length);
      break;
    case USER_DATA_START_CODE:
      user_data(Buf,startcodepos,length);
      break;
    case VIDEO_EDIT_CODE:
      video_edit_code_data(Buf,startcodepos,length);
      break;
    case I_PICTURE_START_CODE:
      curr_pic=prev_pos;				
      I_Picture_Header(Buf,startcodepos,length);
      calc_picture_distance(img);

      if(!img->seq_header_indicate)
      {
        img->B_discard_flag = 1;
        fprintf(stdout,"    I   %3d\t\tDIDSCARD!!\n", img->tr);
        break;
      }

      if (input->check_BBV_flag)
      {
        if (!img->new_sequence_flag)
        {
          if (0x20==profile_id && ((0xFFFF==img->last_pic_bbv_delay && 0xFFFF!=bbv_delay) || (0xFFFF!=img->last_pic_bbv_delay && 0xFFFF==bbv_delay)))
          {
            fprintf(p_sreport, "If one picture's bbv_delay is 0xFFFF, then all pictures's bbv_delay is 0xFFFF.\n");
            bbv_delay = 0xFFFF;
          }
          if (0x20!=profile_id && ((0xFFFFFF==img->last_pic_bbv_delay && 0xFFFFFF!=bbv_delay) || (0xFFFFFF!=img->last_pic_bbv_delay && 0xFFFFFF==bbv_delay)))
          {
            printf("If one picture's bbv_delay is 0xFFFFFF, then all pictures's bbv_delay is 0xFFFFFF in non JiZhun Profile.\n");
            bbv_delay = 0xFFFFFF;
          }
        }

        if ((0x20==profile_id && 0xFFFF==bbv_delay) || (0x20!=profile_id && 0xFFFFFF==bbv_delay))
        {
          pBbv->bbv_mode = 1;
        }
        else
        {
          if (img->new_sequence_flag)
          {
            pBbv->bbv_delay = (float)(bbv_delay/90000.0);
          }
          else if (pBbv->check_continue && (bbv_delay<pBbv->bbv_delay*90000*0.95 || bbv_delay>pBbv->bbv_delay*90000*1.05))
          {
            fprintf(p_sreport,"\nThe bbv_delay of frame %d should be %.3f rather than %.3f.\n", FrameNum, pBbv->bbv_delay, bbv_delay/90000.0);
          }
        }

        img->last_pic_bbv_delay = bbv_delay;
      }

      break; 
    case PB_PICTURE_START_CODE:
      curr_pic=prev_pos;		
      PB_Picture_Header(Buf,startcodepos,length);
      calc_picture_distance(img);

      if(!img->seq_header_indicate)
      {
        img->B_discard_flag = 1;
        if(img->type==P_IMG)
          fprintf(stdout,"    P   %3d\t\tDIDSCARD!!\n", img->tr);
        else
          fprintf(stdout,"    B   %3d\t\tDIDSCARD!!\n", img->tr);
        break;
      }
      if (img->seq_header_indicate==1 && img->type!=B_IMG) {
        img->B_discard_flag = 0;
      }
      if (img->type==B_IMG && img->B_discard_flag==1 && !img->no_forward_reference) {
        fprintf(stdout,"    B   %3d\t\tDIDSCARD!!\n", img->tr);
        break;
      }
     
      if (input->check_BBV_flag)
      {
        if (0x20==profile_id && ((0xFFFF==img->last_pic_bbv_delay && 0xFFFF!=bbv_delay) || (0xFFFF!=img->last_pic_bbv_delay && 0xFFFF==bbv_delay)))
        {
          fprintf(p_sreport, "If one picture's bbv_delay is 0xFFFF, then all pictures's bbv_delay is 0xFFFF.\n");
          bbv_delay = 0xFFFF;
        }
        if (0x20!=profile_id && ((0xFFFFFF==img->last_pic_bbv_delay && 0xFFFFFF!=bbv_delay) || (0xFFFFFF!=img->last_pic_bbv_delay && 0xFFFFFF==bbv_delay)))
        {
          printf("If one picture's bbv_delay is 0xFFFFFF, then all pictures's bbv_delay is 0xFFFFFF in non Baseline Profile.\n");
          bbv_delay = 0xFFFFFF;
        }

        if ((0x20==profile_id && 0xFFFF==bbv_delay) || (0x20!=profile_id && 0xFFFFFF==bbv_delay))
        {
          pBbv->bbv_mode = 1;
        }
        else if (pBbv->check_continue && (bbv_delay<pBbv->bbv_delay*90000*0.95 || bbv_delay>pBbv->bbv_delay*90000*1.05))
        {
          fprintf(p_sreport,"\nThe bbv_delay of frame %d should be %.3f rather than %.3f.\n", FrameNum, pBbv->bbv_delay, bbv_delay/90000.0);
        }

        img->last_pic_bbv_delay = bbv_delay;
      }

      break;  
    case SEQUENCE_END_CODE:
      img->new_sequence_flag = 1; 
      img->sequence_end_flag = 1; 
      free(Buf);
      return EOS;
      break;
    default:
      if ((Buf[startcodepos]>=SLICE_START_CODE_MIN && Buf[startcodepos]<=SLICE_START_CODE_MAX) 
        && ((!img->seq_header_indicate) || (img->type==B_IMG && img->B_discard_flag==1 && !img->no_forward_reference))) {
          break;
      }
      else if (Buf[startcodepos]>=SLICE_START_CODE_MIN) { 
        if ((temp_slice_buf = (char*)calloc (MAX_CODED_FRAME_SIZE , sizeof(char))) == NULL)
          no_mem_exit("GetAnnexbNALU: Buf");
        first_slice_length=length;
        first_slice_startpos=startcodepos;
        memcpy(temp_slice_buf,Buf,length);
        free(Buf);
        return SOP;
      } else {
        printf("Can't find start code");
        free(Buf);
        return EOS;
      }
    }
  }

}

/*
*************************************************************************
* Function:Initializes the parameters for a new frame
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
void init_frame(struct img_par *img, struct inp_par *inp)
{
  static int first_P = TRUE;
  int i,j;

  nextP_tr = nextP_tr_frm;

  if (img->number == 0) // first picture
  {
    nextP_tr=prevP_tr=img->pic_distance; 
  }
  else if(img->type == I_IMG || img->type == P_IMG )
  {
    nextP_tr=img->pic_distance;   

    if(first_P) // first P picture
    {
      first_P = FALSE;
      P_interval=nextP_tr-prevP_tr; 
    }

    if (p_ref && !input->ref_pic_order) 
    {  //ref order
      find_snr(snr,img,p_ref);      // if ref sequence exist
    }
    else
    {
      find_snr(snr,img,p_ref);      // if ref sequence exist
    }

    if(pre_img_type == I_IMG) // I picture
    {
      if(img->new_sequence_flag == 1)
      {
        if(FrameNum!=0)
        {
          printf("min bbv_buffer_size in bitstream is %d\n", ((pminBBSsize>>14) + (pminBBSsize&0x3FFF? 1 : 0)));
          if (!pbbv_mode) 
          {
            printf("min initial bbv_delay(0) time is %.4f(s)\n", (float)pminFsize/pbitrate);
          }			
        }

        if (img->sequence_end_flag) 
        {
          img->sequence_end_flag = 0;
          fprintf(stdout, "Sequence End\n\n");
        }

        if (vec_flag)
        {
          vec_flag = 0;
          fprintf(stdout, "Video Edit Code\n");
        }

        fprintf(stdout, "Sequence Header\n");
        img->new_sequence_flag = 0;
      }
      if(!input->ref_pic_order)  //ref order
        fprintf(stdout,"%3d(I)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s %8d %6d\n",
        FrameNum, pre_img_tr, pre_img_qp,snr->snr_y,snr->snr_u,snr->snr_v,pre_tmp_time,"FRM",StatBitsPtr->prev_frame_bits,StatBitsPtr->prev_emulate_bits);
    }
    else if(pre_img_type == P_IMG) // P pictures
    {
      if (pre_img_types == I_IMG)
      {
        if(img->new_sequence_flag == 1)
        {
          if(FrameNum!=0)
          {
            printf("min bbv_buffer_size in bitstream is %d\n", ((pminBBSsize>>14) + (pminBBSsize&0x3FFF? 1 : 0))); // ITM_r1   
            if (!pbbv_mode) 
            {
              printf("min initial bbv_delay(0) time is %.4f(s)\n", (float)pminFsize/pbitrate);
            }
          }

          if (img->sequence_end_flag) 
          {
            img->sequence_end_flag = 0;
            fprintf(stdout, "Sequence End\n\n");
          } 

          if (vec_flag)
          {
            vec_flag = 0;
            fprintf(stdout, "Video Edit Code\n");
          }

          fprintf(stdout, "Sequence Header\n");
          img->new_sequence_flag = 0;
        }
        if(!input->ref_pic_order) //ref order
          fprintf(stdout,"%3d(I)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s %8d %6d\n",
          FrameNum, pre_img_tr, pre_img_qp,snr->snr_y,snr->snr_u,snr->snr_v,pre_tmp_time, "FRM",StatBitsPtr->prev_frame_bits,StatBitsPtr->prev_emulate_bits);   
      }
      else
      {      
        if(!input->ref_pic_order) // ref order
          fprintf(stdout,"%3d(P)  %3d %5d %7.4f %7.4f %7.4f %5d\t\t%s %8d %6d\n",
          FrameNum, pre_img_tr, pre_img_qp,snr->snr_y,snr->snr_u,snr->snr_v,pre_tmp_time,"FRM",StatBitsPtr->prev_frame_bits,StatBitsPtr->prev_emulate_bits);
      }
    }

    report_frame(snr,pre_tmp_time);

    if (input->output_dec_pic)  // output_dec_pic
      write_prev_Pframe(img, p_out);  // imgY_prev, imgUV_prev -> file
  }

  // allocate memory for frame buffers
  if (img->number == 0) 
  {
    init_global_buffers(inp, img); 
  }

  for(i=0; i<img->max_mb_nr; i++)
  {
    mb_data[i].slice_nr = -1; 
  }

#ifdef MRF_HYU
#ifdef MRF_RAND_P
	  for(i = 0 ; i < img->number_reference_frame ; i++)
		  mref[i] = mref_frm[i];

	  for(i = 0 ; i < img->number_reference_frame ; i++)
		  for(j = 0 ; j < 2 ; j++)      
			  mcef[i][j] = mcef_frm[i][j];
#else
  if(low_delay)
  {
	  for(i = 0 ; i < img->number_reference_frame ; i++)
		  mref[i] = mref_frm[i];

	  for(i = 0 ; i < img->number_reference_frame ; i++)
		  for(j = 0 ; j < 2 ; j++)      
			  mcef[i][j] = mcef_frm[i][j];
  }
  else
  {
	  //luma
	  mref[0] = mref_frm[0]; //mref[ref_index][yuv]height(height/2)][width] ref_index=0,1 for P frame, 
	  mref[1] = mref_frm[1]; 

	  //chroma
	  mcef[0][0] = mcef_frm[0][0];       //mcef[ref_index][uv][height][width]
	  mcef[0][1] = mcef_frm[0][1];
	  mcef[1][0] = mcef_frm[1][0];
	  mcef[1][1] = mcef_frm[1][1];
  }
#endif
#else
  //luma
  mref[0] = mref_frm[0]; //mref[ref_index][yuv]height(height/2)][width] ref_index=0,1 for P frame, 
  mref[1] = mref_frm[1]; 
  
  //chroma
  mcef[0][0] = mcef_frm[0][0];       //mcef[ref_index][uv][height][width]
  mcef[0][1] = mcef_frm[0][1];
  mcef[1][0] = mcef_frm[1][0];
  mcef[1][1] = mcef_frm[1][1];
#endif

  mref_fref[0] = mref_fref_frm[0]; //mref_fref[ref_index][yuv]height(height/2)][width] ref_index=0 for B frame, 
  mref_bref[0] = mref_bref_frm[0];

  mcef_fref[0][0] = 	mcef_fref_frm[0][0]; //mcef_fref[ref_index][uv][height/2][width]
  mcef_fref[0][1] = 	mcef_fref_frm[0][1];
  mcef_bref[0][0] = 	mcef_bref_frm[0][0];
  mcef_bref[0][1] =   mcef_bref_frm[0][1]; 

  imgY = imgY_frm =  current_frame[0];
  imgUV = imgUV_frm =  &current_frame[1];

  img->mv = img->mv_frm;
  refFrArr = refFrArr_frm;

  img->fw_refFrArr = img->fw_refFrArr_frm;
  img->bw_refFrArr = img->bw_refFrArr_frm;

}


void init_frame_buffer()
{

#ifdef MRF_HYU
#ifdef MRF_RAND_P
	int i;   
	for(i = 0 ; i < img->number_reference_frame ; i++)
	{
		mref[i] = mref_frm[i];        

		mcef[i][0] = mcef_frm[i][0];
		mcef[i][1] = mcef_frm[i][1];
	}
#else
	if(low_delay)
	{
		int i, j;   
		for(i = 0 ; i < img->number_reference_frame ; i++)
		{
			mref[i] = mref_frm[i];        

			mcef[i][0] = mcef_frm[i][0];
			mcef[i][1] = mcef_frm[i][1];
		}
	}
	else
	{
		//luma
		mref[0] = mref_frm[0]; //mref[ref_index][yuv]height(height/2)][width] ref_index=0,1 for P frame, 
		mref[1] = mref_frm[1]; 

		//chroma
		mcef[0][0] = mcef_frm[0][0];       //mcef[ref_index][uv][height][width]
		mcef[0][1] = mcef_frm[0][1];
		mcef[1][0] = mcef_frm[1][0];
		mcef[1][1] = mcef_frm[1][1];
	}
#endif
#else
  //luma
  mref[0] = mref_frm[0]; //mref[ref_index][yuv]height(height/2)][width] ref_index=0,1 for P frame, 
  mref[1] = mref_frm[1]; 
  
  //chroma
  mcef[0][0] = mcef_frm[0][0];       //mcef[ref_index][uv][height][width]
  mcef[0][1] = mcef_frm[0][1];
  mcef[1][0] = mcef_frm[1][0];
  mcef[1][1] = mcef_frm[1][1];
#endif

  mref_fref[0] = mref_fref_frm[0]; //mref_fref[ref_index][yuv]height(height/2)][width] ref_index=0 for B frame, 
  mref_bref[0] = mref_bref_frm[0];

  mcef_fref[0][0] = 	mcef_fref_frm[0][0]; //mcef_fref[ref_index][uv][height/2][width]
  mcef_fref[0][1] = 	mcef_fref_frm[0][1];
  mcef_bref[0][0] = 	mcef_bref_frm[0][0];
  mcef_bref[0][1] =   mcef_bref_frm[0][1]; 

  imgY_frm =  current_frame[0];
  imgUV_frm =  &current_frame[1];
}


/*
*************************************************************************
* Function:decodes one picture
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void picture_data(struct img_par *img,struct inp_par *inp)
{
  unsigned char *Buf;
  int startcodepos,length; 
  const int mb_nr = img->current_mb_nr;
  int mb_width = img->width/16;
  int first_slice =1;
  Macroblock *currMB = &mb_data[mb_nr];
  Slice *currSlice = img->currentSlice;     
  int    ByteStartPosition;
  int    i;
  int    new_slice=0;
  Boolean   aec_mb_stuffing_bit;
  img->current_slice_nr = -1;				      
  currentbitoffset = currStream->frame_bitoffset;   
  currStream->frame_bitoffset = 0;               
  if ((Buf = (char*)calloc (MAX_CODED_FRAME_SIZE , sizeof(char))) == NULL) 
    no_mem_exit("GetAnnexbNALU: Buf");	 

  img->cod_counter=-1;

  while (img->current_mb_nr<img->PicSizeInMbs) // loop over macroblocks
  {
    if(img->current_mb_nr%mb_width ==0) 
    {
      if(img->cod_counter<=0)
      {
        if(first_slice)
        {
          SliceHeader(temp_slice_buf,first_slice_startpos,first_slice_length);
          free(temp_slice_buf);
          img->current_slice_nr++;
          img->cod_counter = -1; 
          first_slice=0;
          new_slice=1;
        }
        else 
        {
          if(checkstartcode())
          {
            GetOneUnit(Buf,&startcodepos,&length);
            StatBitsPtr->curr_frame_bits += length;	
            SliceHeader(Buf,startcodepos,length);
            img->current_slice_nr++;
            img->cod_counter = -1; 
            new_slice=1;
          }
          else
            new_slice=0;
        }

        if(input->symbol_mode == AEC && new_slice)
        {

          init_contexts(img);
          AEC_new_slice();
          ByteStartPosition = (currStream->frame_bitoffset)/8;
          if (input->symbol_mode==AEC)
          {
            for (i=0; i<1; i++)
            {
              img->currentSlice->partArr[i].readSyntaxElement = readSyntaxElement_AEC;
              img->currentSlice->partArr[i].bitstream = currStream ;
            }
            currStream = currSlice->partArr[0].bitstream;
          }
          if ((currStream->frame_bitoffset)%8!=0)
            ByteStartPosition++;
          arideco_start_decoding (&img->currentSlice->partArr[0].de_AEC, currStream->streamBuffer, (ByteStartPosition), &(currStream->read_len), img->type);
        }	
      }
    }  //decode slice header   
    mb_data[img->current_mb_nr].slice_nr = img->current_slice_nr;    

#if TRACE
    fprintf(p_trace,"\n*********** Pic: %i (I/P) MB: %i Slice: %i Type %d **********\n", img->tr, img->current_mb_nr, img->current_slice_nr, img->type);
#endif

    start_macroblock(img,inp);
    read_one_macroblock(img,inp);
    decode_one_macroblock(img,inp);
    if(input->symbol_mode==AEC && img->cod_counter<=0/*==-1*/)  
      aec_mb_stuffing_bit = AEC_startcode_follows(img, inp, 1);     
    img->current_mb_nr++;
  }

  free(Buf);

#ifdef IVC_DB
  DeblockFrame (img, imgY, imgUV);
#endif
}


/*
*************************************************************************
* Function:Prepare  frame buffer after frame decoding
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void frame_postprocessing(struct img_par *img)
{
  if((img->number)&&(img->type==I_IMG || img->type == P_IMG))
  {
    nextP_tr_frm = nextP_tr;
  }

  if (img->type==I_IMG || img->type == P_IMG) {
    img->PrevPicDistanceLsb = img->pic_distance;
    img->PicDistanceMsb = img->CurrPicDistanceMsb;
  }
}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
static void rotate_buffer ()
{
  Frame *f;

  f = fb->picbuf[1];
  fb->picbuf[1] = fb->picbuf[0];
  fb->picbuf[0] = f;

  mref[0] = fb->picbuf[0]->mref;
  mcef[0] = fb->picbuf[0]->mcef;
  mref[1] = fb->picbuf[1]->mref;
  mcef[1] = fb->picbuf[1]->mcef;
}


/*
******************************************************************************
*  Function: Determine the MVD's value (1/4 pixel) is legal or not.
*  Input: 
*  Output: 
*  Return:   0: out of the legal mv range; 1: in the legal mv range
*  Attention:
*  Author: 
******************************************************************************
*/
void DecideMvRange()
{
    switch(level_id) {
	case 0x10:
	  Min_V_MV = -512;
	  Max_V_MV =  511;
	  Min_H_MV = -8192;
	  Max_H_MV =  8191;
	  break;
	case 0x20:
		Min_V_MV = -1024;
		Max_V_MV =  1023;
	  Min_H_MV = -8192;
	  Max_H_MV =  8191;
	  break;
	case 0x40:
	   Min_V_MV = -2048;
	   Max_V_MV =  2047;
	  Min_H_MV = -8192;
	  Max_H_MV =  8191;
	  break;
    }
}
