/***************************************************************************************
This software is developed by Peking University, Tsinghua University and Zhejiang University etc. to be submitted to 
ISO/IEC JTC1/SC29/WG11 for evaluation as test platform of internet video coding. 
* Main Contributors include but are not limited as follows:   
Ronggang Wang, Peking University, Shenzhen Graduate School, rgwang@pkusz.edu.cn,   
Xianguo Zhang,   Peking University
Hao Lv,  Peking University, Shenzhen Graduate School
Zhenyu Wang,  Peking University, Shenzhen Graduate School
Xingguo Zhu,  Zhejiang University
Jianwen Chen,  Tsinghua University
Li Zhang,  Peking University
Siwei Ma,  Peking University
Qin Yu, Peking University
etc.
***************************************************************************************/



/*
*************************************************************************************
* File name: 
* Function: 
*
*************************************************************************************
*/
#include <stdlib.h>
#include <string.h>
#include "global.h"
extern const byte QP_SCALE_CR[64] ;

/*********************************************************************************************************/

#define  IClip( Min, Max, Val) (((Val)<(Min))? (Min):(((Val)>(Max))? (Max):(Val)))

byte CLIP_TAB[64] = {
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 2, 2,
	2, 2, 2, 2, 2, 2, 3, 3,
	3, 3, 3, 3, 3, 4, 4, 4,
	5, 5, 5, 6, 6, 6, 7, 7,
	7, 7, 8, 8, 8, 9, 9, 9
} ;

#ifdef IVC_DB

void DeblockMb(ImageParameters *img, byte **imgY, byte ***imgUV, int blk_y, int blk_x) ;
void EdgeLoopX( byte* SrcPtr, int dir, int width, int Chro, int edge );

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void DeblockFrame(ImageParameters *img, byte **imgY, byte ***imgUV)
{
	int       mb_x, mb_y ;
	img->current_mb_nr = -1;    // jlzheng  7.18

	for( mb_y=0 ; mb_y<(img->height>>4) ; mb_y++ )
		for( mb_x=0 ; mb_x<(img->width>>4) ; mb_x++ )
		{
			img->current_mb_nr++;   // jlzheng 7.18
			DeblockMb( img, imgY, imgUV, mb_y, mb_x ) ;
		}
}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void DeblockMb(ImageParameters *img, byte **imgY, byte ***imgUV, int mb_y, int mb_x)
{
	int           EdgeCondition;
	int           dir, edge ;                                                          
	byte          *SrcY, *SrcU = NULL, *SrcV = NULL;
	Macroblock    *MbQ ; 
	const int mb_width = img->width/16;
	const int mb_nr    = img->current_mb_nr;
	Macroblock *currMB = &mb_data[mb_nr];

	SrcY = imgY    [mb_y<<4] + (mb_x<<4) ;    // pointers to source
	if (imgUV != NULL)
	{
		SrcU = imgUV[0][mb_y<<3] + (mb_x<<3) ;
		SrcV = imgUV[1][mb_y<<3] + (mb_x<<3) ;
	}
	MbQ  = &mb_data[mb_y*(img->width>>4) + mb_x] ;     // current Mb

	if (MbQ->lf_disable) return;
	for( dir=0 ; dir<2 ; dir++ )                            // vertical edges, than horicontal edges
	{
		EdgeCondition = (dir && mb_y) || (!dir && mb_x)  ;    // can not filter beyond frame boundaries

		if(dir && mb_y)
		{
			EdgeCondition = (currMB->slice_nr == mb_data[img->current_mb_nr-mb_width].slice_nr)? EdgeCondition:0;  //  can not filter beyond slice boundaries   jlzheng 7.8
		}
		for( edge=0 ; edge<2 ; edge++ )                                            // first 4 vertical strips of 16 pel
		{                                                                          // then  4 horicontal
			if( edge || EdgeCondition )
			{   
				EdgeLoopX( SrcY + (edge<<3)* ((dir)? img->width:1 ), dir, img->width, 0, edge );	

				if ( ( imgUV != NULL ) && ( !edge & 1 ) )
				{
					EdgeLoopX ( SrcU + ( edge << 2 ) * ( ( dir ) ? img->width_cr : 1 ),  dir, img->width_cr, 1, edge ) ;
					EdgeLoopX ( SrcV + ( edge << 2 ) * ( ( dir ) ? img->width_cr : 1 ),  dir, img->width_cr, 1, edge ) ;
				}
			}
		}//end edge
	}//end loop dir
}

/*
*************************************************************************
* Function: EdgeLoopX
* Input:  byte* SrcPtr,int QP, int dir,int width,int Chro
* Output: void
* Return: void
*************************************************************************
*/

void EdgeLoopX ( byte* SrcPtr, int dir, int width, int Chro,int edge ) 
{

	int     pel , PtrInc;
	int     inc, inc2, inc3, inc4;
	int     Axxxelta, LeftD, RightD ;
	int     L3, L2, L1, L0, R0, R1, R2, R3;
	int     fs = 0; //fs stands for filtering strength.  The larger fs is, the stronger filter is applied.
	int     flatedge = 0, edge_th = 0;

	PtrInc  = dir ?  1 : width ;
	inc     = dir ?  width : 1 ;
	inc2    = inc << 1 ;
	inc3    = inc + inc2;
	inc4    = inc + inc3;

	for ( pel = 0; pel < 16 ; pel++ )
	{
		L3  = SrcPtr[-inc4] ;
		L2  = SrcPtr[-inc3] ; 
		L1  = SrcPtr[-inc2] ;
		L0  = SrcPtr[-inc ] ;
		R0  = SrcPtr[    0] ;
		R1  = SrcPtr[ inc ] ;
		R2  = SrcPtr[ inc2] ;
		R3  = SrcPtr[ inc3] ;

		Axxxelta = abs ( R0 - L0 ) ;
		LeftD = abs ( L0 - L1 ) ;
		RightD = abs ( R0 - R1 ) ;

		edge_th = MIN(3, Beta);
		flatedge = LeftD < edge_th && RightD < edge_th;

		if ( LeftD < Beta && RightD < Beta && Axxxelta > LeftD && Axxxelta > RightD && Axxxelta < Alpha)
			if ( abs(L2 - L0) < Beta && abs(R2 - R0) < Beta)
				fs = 2;
			else
				fs = 1;
		else
			fs = 0;

		if (fs && Chro && !flatedge)
		{
			fs = 1;
		}

		if ((fs==2) && flatedge && !edge && !Chro && abs(R3 - R0) < Beta && abs(L3 - L0) < Beta)
		{
			fs = 3;
		}

		switch ( fs )
		{
		case 3:
			SrcPtr[-inc3] = (( 4 * ( L0 - R0) + 5 * ( R0 - L2 ) + 4 ) >> 3) + L2; //L2
			SrcPtr[-inc2] = (( 16 * ( R0 - L1 ) + 6 * ( L2 - R0 ) + 7* ( L0 - R0 ) + 8 ) >> 4) + L1; //L1
			SrcPtr[-inc]  = (( 9 * ( L2 - R0) + 6 * ( R2 - L0 ) + 17 * ( R0 - L0 ) + 16 ) >> 5) + L0; //L0
			SrcPtr[0]     = (( 9 * ( R2 - L0) + 6 * ( L2 - R0 ) + 17 * ( L0 - R0 ) + 16 ) >> 5) + R0; //R0
			SrcPtr[inc]   = (( 16 * ( L0 - R1 ) + 6 * ( R2 - L0 ) + 7* ( R0 - L0 ) + 8 ) >> 4) + R1; //R1
			SrcPtr[inc2]  = (( 4 * ( R0 - L0) + 5 * ( L0 - R2 ) + 4 ) >> 3) + R2; //R2
			break;

		case 2:
			SrcPtr[-inc2] = (((L2 - R0) * 3 + (R0 - L1) * 8 + (L0 - R0) * 4 +8)>> 4) + L1;//L1
			SrcPtr[-inc ] = (((L2 - R0) + (L1 - R0) * 4 + (R1 - L0) + (R0 - L0) * 9 + 8)>> 4) + L0;//L0
			SrcPtr[  0  ] = (((R2 - L0) + (R1 - L0) * 4 + (L1 - R0) + (L0 - R0) * 9 + 8)>> 4) + R0;//R0
			SrcPtr[ inc ] = (((R2 - L0) * 3 + (L0 - R1) * 8 + (R0 - L0) * 4 +8) >> 4) + R1;//R1
			break;

		case 1:
			SrcPtr[0] =  ((L0 - R0 + 2)>>2) + R0; //R0
			SrcPtr[-inc] = ((R0 - L0 + 2)>>2) + L0; //L0
			break;

		default:
			break;
		}

		SrcPtr += PtrInc ;    // Next row or column
		pel    += Chro ;	
	}
}
#endif

